﻿/// @file config.cpp
///
/// @brief config
///
#include "config.h"
#include <wx/filename.h>
#include <wx/fileconf.h>

Config gConfig;

Config::Config()
{
	ini_file = _T("");

	// default value
	mFilePath = _T("");
	mMemoryCash = false;
	mServerHost = _T("localhost");
	mServerPort = 10200;
	mPrintDelay = 16;
	mDataReverse = true;
	mDensity = COLOR_DENSITY_MAX;
}

Config::~Config()
{
}

void Config::SetFileName(const wxString &file)
{
	ini_file = file;
}

void Config::Load()
{
	if (ini_file.IsEmpty()) return;

	// load ini file
	wxFileConfig *ini = new wxFileConfig(wxEmptyString,wxEmptyString,ini_file,wxEmptyString
		,wxCONFIG_USE_LOCAL_FILE | wxCONFIG_USE_RELATIVE_PATH | wxCONFIG_USE_NO_ESCAPE_CHARACTERS);

	int i;
	long l;
	wxString s;
	ini->Read(_T("Path"), &mFilePath, mFilePath);
	ini->Read(_T("MemoryCash"), &mMemoryCash, mMemoryCash);
	ini->Read(_T("ServerHost"), &mServerHost, mServerHost);
	ini->Read(_T("ServerPort"), &l, mServerPort);
	if (0 < l && l <= 65535) mServerPort = l;
	ini->Read(_T("PrintDelay"), &i, mPrintDelay);
	if (0 <= i && i <= 10000) mPrintDelay = i; 
	ini->Read(_T("Density"), &i, mDensity);
	if (0 <= i && i <= COLOR_DENSITY_MAX) mDensity = i; 
	ini->Read(_T("DataReversed"), &mDataReverse, mDataReverse);
	for(int i=0; i<MAX_RECENT_FILES; i++) {
		wxString sval;
		ini->Read(wxString::Format(_T("Recent%d"), i), &sval);
		if (!sval.IsEmpty()) {
			mRecentFiles.Add(sval);
		}
	}
	delete ini;
}

void Config::Load(const wxString &file)
{
	SetFileName(file);
	Load();
}

void Config::Save()
{
	if (ini_file.IsEmpty()) return;

	// save ini file
	wxFileConfig *ini = new wxFileConfig(wxEmptyString,wxEmptyString,ini_file,wxEmptyString
		,wxCONFIG_USE_LOCAL_FILE | wxCONFIG_USE_RELATIVE_PATH | wxCONFIG_USE_NO_ESCAPE_CHARACTERS);
	ini->Write(_T("Path"), mFilePath);
	ini->Write(_T("MemoryCash"), mMemoryCash);
	ini->Write(_T("ServerHost"), mServerHost);
	ini->Write(_T("ServerPort"), mServerPort);
	ini->Write(_T("PrintDelay"), mPrintDelay);
	ini->Write(_T("Density"), mDensity);
	ini->Write(_T("DataReversed"), mDataReverse);
	for(int i=0,row=0; row<MAX_RECENT_FILES && i<(int)mRecentFiles.Count(); i++) {
		wxString sval = mRecentFiles.Item(i);
		if (sval.IsEmpty()) continue;
		ini->Write(wxString::Format(_T("Recent%d"), row), sval);
		row++;
	}
	// write
	delete ini;
}

void Config::AddRecentFile(const wxString &val)
{
	wxFileName fpath = wxFileName::FileName(val);
	mFilePath = fpath.GetPath(wxPATH_GET_SEPARATOR);
	// 同じファイルがあるか
	int pos = mRecentFiles.Index(fpath.GetFullPath());
	if (pos >= 0) {
		// 消す
		mRecentFiles.RemoveAt(pos);
	}
	// 追加
	mRecentFiles.Insert(fpath.GetFullPath(), 0);
	// 10を超える分は消す
	if (mRecentFiles.Count() > MAX_RECENT_FILES) {
		mRecentFiles.RemoveAt(MAX_RECENT_FILES);
	}
}

wxString &Config::GetRecentFile()
{
	return mRecentFiles[0];
}

void Config::GetRecentFiles(wxArrayString &vals)
{
	vals = mRecentFiles;
}
