﻿/// @file server_setting.cpp
///
/// @brief MPPRINTER サーバーダイアログ制御
///
/// @date 2013.11.27 Create
///
/// @author Copyright (c) 2013, Sasaji. All rights reserved.
///
#include "server_setting.h"

// Attach Event
BEGIN_EVENT_TABLE(ServerBox, wxDialog)
	// event
	EVT_BUTTON(wxID_OK, ServerBox::OnButton)
END_EVENT_TABLE()

ServerBox::ServerBox(wxWindow* parent, wxWindowID id)
	: wxDialog(parent, id, _("Server settings"), wxDefaultPosition, wxDefaultSize, wxDEFAULT_DIALOG_STYLE)
{
	wxSizerFlags flags = wxSizerFlags().Expand().Border(wxALL, 4);

	wxBoxSizer *szrAll = new wxBoxSizer(wxVERTICAL);
	wxFlexGridSizer *gbox = new wxFlexGridSizer(
#ifdef USE_DELAY_SETTING
		4,
#else
		3,
#endif
		3,0,0);

	txtHost = new wxTextCtrl(this, IDC_TXT_HOST);
	gbox->Add(new wxStaticText(this, wxID_ANY, _("Hostname:")), flags);
	gbox->Add(txtHost, flags);
	gbox->Add(new wxStaticText(this, wxID_ANY, wxT("")), flags);

	txtPort = new wxTextCtrl(this, IDC_TXT_PORT);
	gbox->Add(new wxStaticText(this, wxID_ANY, _("Port:")), flags);
	gbox->Add(txtPort, flags);
	gbox->Add(new wxStaticText(this, wxID_ANY, wxT("")), flags);

	chkReverse = new wxCheckBox(this, IDC_CHK_REVERSE, _("Reverse Data"));
	gbox->Add(new wxStaticText(this, wxID_ANY, wxT("")), flags);
	gbox->Add(chkReverse, flags);
	gbox->Add(new wxStaticText(this, wxID_ANY, wxT("")), flags);

#ifdef USE_DELAY_SETTING
	spinDelay = new wxSpinCtrl(this, IDC_SPIN_DELAY);
	gbox->Add(new wxStaticText(this, wxID_ANY, _("Print Delay:")), flags);
	gbox->Add(spinDelay, flags);
	gbox->Add(new wxStaticText(this, wxID_ANY, _("ms.")), flags);
#endif

	szrAll->Add(gbox, flags);

	wxSizer *szrButtons = CreateButtonSizer(wxOK|wxCANCEL);
	szrAll->Add(szrButtons, flags);

	SetSizerAndFit(szrAll);
}

void ServerBox::init_dialog()
{
	txtHost->SetValue(mHost);
	txtPort->SetValue(wxString::Format(wxT("%ld"), mPort));
	chkReverse->SetValue(mReverse);
#ifdef USE_DELAY_SETTING
	spinDelay->SetRange(0, 10000);
	spinDelay->SetValue(mDelay);
#endif
}

void ServerBox::term_dialog()
{
	mHost = txtHost->GetValue();
	long l;
	txtPort->GetValue().ToLong(&l);
	if (0 < l && l <= 65535) mPort = l;
	mReverse = chkReverse->GetValue();
#ifdef USE_DELAY_SETTING
	int i = spinDelay->GetValue();
	if (0 <= i && i <= 10000) mDelay = i;
#endif
}

void ServerBox::OnButton(wxCommandEvent& event)
{
	wxDialog::EndDialog(wxID_OK);
}

int ServerBox::ShowModal()
{
	// initialize
	init_dialog();
	int rc = wxDialog::ShowModal();
	if (rc == wxID_OK) {
		// update
		term_dialog();
	}
	return rc;
}

